@@ -9,6 +9,9 @@ import org.jruby.exceptions.RaiseException;
 import org.jruby.runtime.ThreadContext;
 import org.jruby.util.ByteList;
 
+import java.io.IOException;
+import java.io.OutputStream;
+
 /**
  * An encoder that reads from the given source and outputs its representation
  * to another ByteList. The source string is fully checked for UTF-8 validity,
@@ -17,10 +20,12 @@ import org.jruby.util.ByteList;
 final class StringEncoder extends ByteListTranscoder {
     private final boolean asciiOnly, scriptSafe;
 
+    private OutputStream out;
+
     // Escaped characters will reuse this array, to avoid new allocations
     // or appending them byte-by-byte
     private final byte[] aux =
-        new byte[] {/* First unicode character */
+        new byte[] {/* First Unicode character */
                     '\\', 'u', 0, 0, 0, 0,
                     /* Second unicode character (for surrogate pairs) */
                     '\\', 'u', 0, 0, 0, 0,
@@ -37,23 +42,42 @@ final class StringEncoder extends ByteListTranscoder {
             new byte[] {'0', '1', '2', '3', '4', '5', '6', '7',
                         '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'};
 
-    StringEncoder(ThreadContext context, boolean asciiOnly, boolean scriptSafe) {
-        super(context);
+    StringEncoder(boolean asciiOnly, boolean scriptSafe) {
         this.asciiOnly = asciiOnly;
         this.scriptSafe = scriptSafe;
     }
 
-    void encode(ByteList src, ByteList out) {
-        init(src, out);
+    void encode(ThreadContext context, ByteList src, OutputStream out) throws IOException {
+        init(src);
+        this.out = out;
         append('"');
         while (hasNext()) {
-            handleChar(readUtf8Char());
+            handleChar(readUtf8Char(context));
         }
         quoteStop(pos);
         append('"');
     }
 
-    private void handleChar(int c) {
+    void encodeASCII(ThreadContext context, ByteList src, OutputStream out) throws IOException {
+        init(src);
+        this.out = out;
+        append('"');
+        while (hasNext()) {
+            handleChar(readASCIIChar());
+        }
+        quoteStop(pos);
+        append('"');
+    }
+
+    protected void append(int b) throws IOException {
+        out.write(b);
+    }
+
+    protected void append(byte[] origin, int start, int length) throws IOException {
+        out.write(origin, start, length);
+    }
+
+    private void handleChar(int c) throws IOException {
         switch (c) {
         case '"':
         case '\\':
@@ -97,13 +121,13 @@ final class StringEncoder extends ByteListTranscoder {
         }
     }
 
-    private void escapeChar(char c) {
+    private void escapeChar(char c) throws IOException {
         quoteStop(charStart);
         aux[ESCAPE_CHAR_OFFSET + 1] = (byte)c;
         append(aux, ESCAPE_CHAR_OFFSET, 2);
     }
 
-    private void escapeUtf8Char(int codePoint) {
+    private void escapeUtf8Char(int codePoint) throws IOException {
         int numChars = Character.toChars(codePoint, utf16, 0);
         escapeCodeUnit(utf16[0], ESCAPE_UNI1_OFFSET + 2);
         if (numChars > 1) escapeCodeUnit(utf16[1], ESCAPE_UNI2_OFFSET + 2);
@@ -117,8 +141,7 @@ final class StringEncoder extends ByteListTranscoder {
     }
 
     @Override
-    protected RaiseException invalidUtf8() {
-         return Utils.newException(context, Utils.M_GENERATOR_ERROR,
-                 "source sequence is illegal/malformed utf-8");
+    protected RaiseException invalidUtf8(ThreadContext context) {
+        return Utils.newException(context, Utils.M_GENERATOR_ERROR, "source sequence is illegal/malformed utf-8");
     }
 }
